/*
 *   comparer.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "gripho.h"
/*
    comparer <fbase> <fnew> <ftrim> <cutoff> <range> <mul>
       fbase   file di riferimento per il test
       fnew    file da verificare
       ftrim   file che verra' utilizzato per mascherare le due img
       cutoff  val al di sotto del quale le differenze tra comp dei pixel
               verranno considerate nulle
       range   lato del crocere che viene testato per amplificare la diff
               del pixel
       mul     valore massimo di amplificazione

    */

#define DEMULTI 1000.0

void my_error(char *frm,char *s,int ext)
{
  if (s) 
    fprintf(stderr,frm,s);
  else
    fprintf(stderr,frm);
  exit (ext);
}

int trim_layer(gr_layer *ed,gr_layer *er)
{
  int i,e,a,c,x,y,nchan;
  
  x = ed->x;
  y = ed->y;
  nchan = ed->nchan;
  if (x != er->x || y != er->y || nchan != er->nchan) 
    return (0);
  c = (1 << ed->depth) / 2;
  for (i = 0 ; i < y ; i++)
    for (e = 0 ; e < x ; e++)
      for (a = 0 ; a < nchan ; a++) 
	if (((unsigned char *)er->dt[a][i])[e] < c)
	  ((unsigned char *)ed->dt[a][i])[e] = 0;
  return (1);
}

int diff_layer(gr_layer *ed,gr_layer *er,int cutoff)
{
  int i,e,a,c,x,y,nchan;
  unsigned char *p;
  x = ed->x;
  y = ed->y;
  nchan = ed->nchan;
  if (x != er->x || y != er->y || nchan != er->nchan) 
    return (0);
  for (i = 0 ; i < y ; i++)
    for (e = 0 ; e < x ; e++)
      for (a = 0 ; a < nchan ; a++) {
	p = ((unsigned char *)ed->dt[a][i]);
	c = ( p[e] - ((unsigned char *)er->dt[a][i])[e]);
	c = (c < 0 ? -c : c);
	if (c > cutoff)
	  p[e] = c;
	else
	  p[e] = 0;
      }

  return (1);
}


unsigned long maxim_layer(gr_layer *ly)
{
  int i,e,a,c,m,x,y,nchan;

  x = ly->x;
  y = ly->y;
  nchan = ly->nchan;
  for (i = 0 ; i < y ; i++)
    for (e = 0 ; e < x ; e++) {
      m = 0;
      for (a = 0 ; a < nchan-1 ; a++) {
	c = (int)((unsigned char *)ly->dt[a][i])[e];
	if (m < c)
	  m = c;
      }
      ((unsigned char *)ly->dt[a][i])[e] = (unsigned char)m;
    }
  return (1);
}

/*
   logica adottata:
      per ogni pixel il viene presa il chan con la differenza maggiore
      tale valore viene quadrato
      si prende la media del crocere e a seconda della differenza ta tale
         media e il valore del pixel si moltiplica il quadrato da 1 a mul
         volte se sono uguali
      */
unsigned long coeff_layer(gr_layer *ly,int chan,int range,int imul)
{
  int i,e,a,x,y,nchan,xi,yi,xmin,xmax,ymin,ymax;
  double c,cc,co,cco,sum,ctp,mul;
  x = ly->x;
  y = ly->y;
  nchan = ly->nchan;
  a = chan;
  mul = (double)imul;

  if (range == 1) {
    for (i = 0 ; i < y ; i++)
      for (e = 0 ; e < x ; e++) {
	c = (double)((unsigned char *)ly->dt[a][i])[e];
	c = c * c;
	co += c;
      }
  }
  else {
    co = 0.0;
    for (i = 0 ; i < y ; i++) {
      for (e = 0 ; e < x ; e++) {
	cc = (double)((unsigned char *)ly->dt[a][i])[e];
	if (cc == 0.0)
	  continue;
	c = cc * cc;
	ymin = ((i-range) < 0 ? 0 : (i-range));
	ymax = ((i+range) > y ? y : (i+range));
	xmin = ((e-range) < 0 ? 0 : (e-range));
	xmax = ((e+range) > x ? x : (e+range));
	sum = 0;
	ctp = 0;
	for (yi = ymin ; yi < ymax ; yi++) 
	  for (xi = xmin ; xi < xmax ; xi++) {
	    sum += (double)((unsigned char *)ly->dt[a][yi])[xi];
	    ctp++;
	  }
	sum /= ctp;
	if (sum > cc)
	  co += (c * mul);
	else
	  co += c * (((mul - 1.0) * ((double)sum / (double)cc)) + 1);
      }
    }
  }
  cco = (double)co / DEMULTI;
  if (cco > 4000000000.0)
    cco = 4000000000.0;
  return ((unsigned long)cco);
}

int main(int argc,char **argv)
{
  FILE *fbas,*fnew,*ftrm;
  gripho *gbas,*gnew,*gtrm;
  int cutoff,mul,range;

  if (argc != 7) {
    fprintf(stderr,"\n    compview <fbase> <fnew> <ftrim> <cutoff> <range> <mul>\n\
       fbase   file di riferimento per il test\n\
       fnew    file da verificare\n\
       ftrim   file che verra' utilizzato per mascherare le due img\n\
       cutoff  val al di sotto del quale le differenze tra comp dei pixel\n\
               verranno considerate nulle\n\
       range   lato del crocere che viene testato per amplificare la diff\n\
               del pixel\n\
       mul     valore massimo di amplificazione\n\n");
    exit(10);
  }

  if ((fbas = fopen(argv[1],"r")) == NULL)
    my_error("open base error (%s)\n",argv[1],1);
  if ((fnew = fopen(argv[2],"r")) == NULL)
    my_error("open new error (%s)\n",argv[2],1);
  if ((ftrm = fopen(argv[3],"r")) == NULL)
    my_error("open trim error (%s)\n",argv[3],1);
  if ((gbas = import_pnm(fbas)) == NULL)
    my_error("import base error (%s)\n",argv[1],1);
  if ((gnew = import_pnm(fnew)) == NULL)
    my_error("import new error (%s)\n",argv[2],1);
  if ((gtrm = import_pnm(ftrm)) == NULL)
    my_error("import trim error (%s)\n",argv[3],1);
  cutoff = atoi(argv[4]);
  range = atoi(argv[5]);
  mul = atoi(argv[6]);
  if (!diff_layer(gnew->lays[0],gbas->lays[0],cutoff))
    my_error("diff error\n",NULL,2);
  if (!trim_layer(gnew->lays[0],gtrm->lays[0]))
    my_error("trim error\n",NULL,2);
  img_addchan(gnew,GR_C_ALPHA,255);
  maxim_layer(gnew->lays[0]);
  printf("Coefficente: %ld\n\n",coeff_layer(gnew->lays[0],chan_getnumid(gnew->schan,GR_C_ALPHA),range,mul));
  

 
  /*  explay_pnm(stdout,gnew->lays[0]);*/
  
  exit(0);
}
