/*
 *   fusepnm.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void usage (char *s,int ext) {
  printf("%s <fileimg> <filemsk> <filetile> <fileout>\n",s);
  exit(ext);
}

int getdim(char *s,int *x,int *y)
{
  int i;

  for (i = 0 ; s[i] != '\0' ; i++) 
    if (s[i] == ' ' || s[i] == '\t')
      break;
  if (s[i] == '\0')
    return (0);
  s[i++] = '\0';
  *x = atoi(s);
  for (; s[i] != '\0' ; i++) 
    if (s[i] != ' ' && s[i] != '\t')
      break;
  if (s[i] == '\0')
    return (0);
  *y = atoi(&(s[i]));
  if (*x < 0 || *y < 0)
    return (0);

  return (1);
}
/*
 *  MAIN
 */

int main (int argc,char **argv)
{
  int i,e,c,intile;
  int xim,yim,xms,yms,xti,yti,comp,trasp,len;
  FILE *fim,*fms,*fti,*fot;
  char bf[512];
  unsigned char *tile;
  double fcomp,ftrasp,ftile,fres,ffraz;
  
  for (i = 1 ; i < argc ; i++) {
    if (argv[i][0] != '-')
      break;
    else
      usage(argv[0],1);
  }
  if (argc - i != 4)
    usage(argv[0],2);
  if ((fim = fopen(argv[i],"r")) == NULL)
    usage(argv[0],3);
  if ((fms = fopen(argv[i+1],"r")) == NULL)
    usage(argv[0],4);
  if ((fti = fopen(argv[i+2],"r")) == NULL)
    usage(argv[0],5);
  if ((fot = fopen(argv[i+3],"w")) == NULL)
    usage(argv[0],6);

  if (fgets(bf,512,fim) == NULL || strncmp(bf,"P6",2))
    exit(7);
  if (fgets(bf,512,fms) == NULL || strncmp(bf,"P5",2))
    exit(8);
  if (fgets(bf,512,fti) == NULL || strncmp(bf,"P6",2))
    exit(9);
  if (fgets(bf,512,fim) == NULL)
    exit(10);
  if (fgets(bf,512,fms) == NULL)
    exit(11);
  if (fgets(bf,512,fti) == NULL)
    exit(11);
  if (fgets(bf,512,fim) == NULL || !getdim(bf,&xim,&yim))
    exit(12);
  if (fgets(bf,512,fms) == NULL || !getdim(bf,&xms,&yms))
    exit(13);
  if (fgets(bf,512,fti) == NULL || !getdim(bf,&xti,&yti))
    exit(14);
  if (xim != xms || yim != yms)
    exit(15);
  if (fgets(bf,512,fim) == NULL || strncmp(bf,"255",3))
    exit(16);
  if (fgets(bf,512,fms) == NULL || strncmp(bf,"255",3))
    exit(17);
  if (fgets(bf,512,fti) == NULL || strncmp(bf,"255",3)) 
    exit(18);
  if ((tile = malloc(xti*yti*3)) == NULL)
    exit(19);
  len = xti * yti * 3;
  for (i = 0 ; i < len ; i++)
    if ((intile = fgetc(fti)) == EOF)
      exit(20);
    else
      tile[i] = intile;
  fprintf(fot,"P6\n#create with fusepnm\n%d %d\n255\n",xim,yim);
  
  for (e = 0 ; e < yim ; e++) {
    for (i = 0 ; i < xim ; i++) {
      trasp = fgetc(fms);
      if (trasp == EOF)
	exit(21);
      ftrasp = (double)trasp;

      for (c = 0 ; c < 3 ; c++) {
	if ((comp = fgetc(fim)) == EOF)
	  exit(22);
	fcomp = (double)comp;
	ftile = (double)tile[(xti * 3 * (e%yti)) + (3 * (i%xti)) + c ];
        ffraz = ftrasp / 255.0;
	fres = fcomp * ffraz + ftile * (1.0 - ffraz);
	comp = (int)fres;
	if (comp < 0)
	  comp = 0;
	else if (comp > 255)
	  comp = 255;
	fputc(comp,fot);
      }
    }
  }
  fclose(fim);
  fclose(fms);
  fclose(fot);
  exit(0);
  return (0);
}
