/*
 *   gr_gripho.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gripho.h"

void lay_info(gr_layer *ly) {
  fprintf(stderr,"tipo di layer: %d\n",ly->type);
  fprintf(stderr,"x: %d  y: %d\n",ly->x, ly->y);
  fprintf(stderr,"dx: %d  dy: %d\n",ly->dx, ly->dy);
  fprintf(stderr,"nchan %d\n",ly->nchan);
  fprintf(stderr,"schan [%s]\n",ly->schan);
  fprintf(stderr,"depth %d\n",ly->depth);
  fprintf(stderr,"trans %d\n",ly->trans);
  fprintf(stderr,"active %d\n",ly->active);
  fprintf(stderr,"com [%s]\n",ly->com);
}

int lay_destroy(gr_layer **ly)
{
  gr_free_data(&((*ly)->dt),(*ly)->nchan,(*ly)->y);
  free(*ly);
  *ly = NULL;

  return (TRUE);
}

gr_layer *lay_create(int x, int y, int nchan, char *schan,
		     int depth, int trans, int active,  char *com)
{
  gr_layer *ly;
  int size;

  if ((ly = malloc(sizeof(gr_layer))) == NULL)
    return NULL;
  ly->x = x;
  ly->y = y;
  ly->nchan = nchan;
  strncpy(ly->schan,schan,31);
  ly->schan[31] = '\0';
  strncpy(ly->com,com,GR_LENCOM-1);
  ly->com[GR_LENCOM-1] = '\0';
  ly->depth = depth;
  ly->trans = trans;
  ly->active = active;
  
  size = nbyte(depth);
  
  /* area dati   dt[chan][raw][col] */
  if ((ly->dt = gr_alloc_data(nchan,size,x,y)) == NULL) {
    free(ly);
    return NULL;
  }
  
  return (ly);
}

void lay_copydata(gr_layer *new,gr_layer *old)
{
  int a,x,y,nb;

  nb = nbyte(new->depth);

  for (a = 0 ; a<new->nchan ; a++) 
    for (y = 0 ; y<new->y ; y++)
      for (x = 0 ; x<new->x ; x++) {
	chan_setpix(new->dt[a][y],x,nb,
		    chan_getpix(old->dt[a][y],x,nb));
      }
}

void lay_copyheader(gr_layer *new,gr_layer *old)
{
  new->x = old->x;
  new->y = old->y;
  new->nchan = old->nchan;
  strncpy(new->schan,old->schan,31);
  new->schan[31] = '\0';
  strncpy(new->com,old->com,GR_LENCOM-1);
  new->com[GR_LENCOM-1] = '\0';
  new->depth = old->depth;
  new->trans = old->trans;
  new->active = old->active;
}

gr_layer *lay_copy(gr_layer *old)
{
  gr_layer *ly;
  int size;

  if ((ly = malloc(sizeof(gr_layer))) == NULL)
    return NULL;
  lay_copyheader(ly,old);
  size = nbyte(old->depth);
  /* area dati   dt[chan][raw][col] */
  if ((ly->dt = gr_alloc_data(old->nchan,size,old->x,old->y)) == NULL) {
    free(ly);
    return NULL;
  }
  lay_copydata(ly,old);

  return (ly);
}

/* sx e sy sono (oldx / newx) e (oldy / newy) */
unsigned long chan_getpixscaled(void ***dt,int xx,int yy,int chan,
				double sx,double sy,int nb) 
{
  double d[2][2], f[2][2], sum = 0.0;
  double t[2][2], mul[2];
  int i, e, a, c[2][2] = {{0,0},{0,0}},inp[2], par[2];
  
  inp[0] = xx;  inp[1] = yy;
  mul[0] = sx;  mul[1] = sy;
  DBG3("DENTRO xx %d yy %d\n",xx,yy);
  for (e = 0 ; e< 2 ; e++) {
    for (i = 0 ; i< 2 ; i++) {
      t[e][i] = ((double)(inp[e]+i)) * mul[e];
      f[e][i] = modf(t[e][i],&(d[e][i]));
      if (ZERO(d[e][i]-t[e][i]) == 0.0) 
	d[e][i] = ATTRACT(t[e][i]);	
      else 
	c[e][i] = 1;
      DBG5("[%d][%d] T= %f c %d\n",e,i,t[e][i],c[e][i]);
    }
  }
  for (e = 0 ; e< 2 ; e++) {
    if (c[e][0]) {
      if (c[e][1] && d[e][0] == d[e][1]) {
	f[e][0] = f[e][1] - f[e][0];
	c[e][1] = 0;
      }
      else
	f[e][0] = 1.0 - f[e][0];
    }
  }
  /* struttura:
       1 | 2 | 3        1,3,7,9   x e y parziali
      ---+---+---       2,8       y parziale
       4 | 5 | 6        4,6       x parziale
      ---+---+---       5         tutti i pix in 5 sono 
       7 | 8 | 9                  conteggiati interamente
   */
  /* 1,3,7,9 */
  for (i = 0 ; i < 2 ; i++)
    for (e = 0 ; e < 2 ; e++)
      if (c[0][i] && c[1][e]) {
	DBG3("[1,3,7,9] x %d y %d\n",((int)d[0][i]),((int)d[1][e])); 
	sum += (((double)
		chan_getpix(dt[chan][((int)d[1][e])],((int)d[0][i]),nb))
		* f[0][i] * f[1][e]);
      }
  /* 2,8,4,6 */
  for (e = 0 ; e < 2 ; e++)
    for (i = 0 ; i < 2 ; i++)
      if (c[e][i]) 
	for (a = ((int)d[1-e][0])+c[1-e][0] ; a < ((int)d[1-e][1]) ; a++) {
	  par[e] = ((int)d[e][i])+i;
	  par[1-e] = a;
          DBG3("[2,8,4,6] x %d y %d\n",par[1],par[0]); 
	  sum += (((double)chan_getpix(dt[chan][par[1]],par[0],nb))
		  * f[e][i]);
	}
  for (e = ((int)d[1][0])+c[1][0] ; e < ((int)d[1][1]) ; e++) 
    for (i = ((int)d[0][0])+c[0][0] ; i < ((int)d[0][1]) ; i++) {
      /* fprintf (stderr,"X %d Y %d\n",i,e); */
      DBG3("[5] x %d y %d\n",i,e); 
      sum = sum + ((double) chan_getpix(dt[chan][e],i,nb));
    }
  /* div = (xf[0]+xf[1]+x[1]-x[0])*(yf[0]+yf[1]+y[1]-y[0]);*/
  /*  fprintf(stderr,"chan %d yy %d xx %d sum %f div %f ris %f\n",chan,yy,xx,sum,div,sum/div); */
  sum = sum / ((t[0][1] - t[0][0]) * (t[1][1] - t[1][0]));

  DBG3("FUORI (%f)(%f)\n",sum,((t[0][1] - t[0][0]) * (t[1][1] - t[1][0])));
  
  return (unsigned long)sum;
}


gr_layer *lay_scalenew(gr_layer *old,int x,int y)
{
  gr_layer *ly;
  int a,i,e,nb;
  double sx,sy;
  void ***dt,***newdt;

  if ((ly = malloc(sizeof(gr_layer))) == NULL)
    return NULL;
  lay_copyheader(ly,old);
  ly->x = x;
  ly->y = y;
  nb = nbyte(old->depth);
  /* area dati   dt[chan][raw][col] */
  if ((ly->dt = gr_alloc_data(old->nchan,nb,x,y)) == NULL) {
    free(ly);
    return NULL;
  }
  
  sx = ((double)old->x-1) / ((double)x);
  sy = ((double)old->y-1) / ((double)y);
  DBG7("zagor %d %d %d %d %f %f\n",old->x,x,old->y,y,sx,sy); 
  dt = old->dt;
  newdt = ly->dt;
  for (a = 0 ; a < old->nchan ; a++) {
    for (e = 0 ; e < y ; e++) {
      for (i = 0 ; i< x ; i++) {
	/*	if (e < 7 && i < 10) */
	  chan_setpix(newdt[a][e],i,nb,
		      chan_getpixscaled(dt,i,e,a,sx,sy,nb));
	       
      }
    }
  }

  return (ly);
}
	
int img_lastlay(gripho *gr) 
{
  return (gr->nlay-1);
}

int img_remlayer(gripho *gr, int idx) 
{
  int i;

  lay_destroy(&(gr->lays[idx]));
  for (i = idx ; i < gr->nlay-1 ; i++)
    gr->lays[i] = gr->lays[i+1];
  gr->nlay--;

  return TRUE;
}
int img_addlayer(gripho *gr,gr_layer *ly,int type,int dx, int dy) 
{
  /* PIPPO qui andra' check channel coerenti e resort degli stessi */
  if ((gr->lays = realloc(gr->lays,
			  sizeof(gr_layer *) * (gr->nlay + 1))) == NULL)
    return (-1);
  ly->type = type;
  ly->dx   = dx;
  ly->dy   = dy;
  gr->lays[gr->nlay] = ly;
  gr->nlay++;

  return (gr->nlay-1);
}

gripho *new_img(int x, int y, int nchan, char *schan, int depth, char *com) 
{
  gripho *gr;
  gr_layer *ly;
  if ((gr = malloc(sizeof(gripho))) != NULL) {
    if ((ly = lay_create(x,y,nchan,schan,depth,
		      GR_TRANS_OPAC,TRUE,"background")) != NULL) {
      gr->nlay = 0;
      if (img_addlayer(gr,ly,GR_LAYTYP_NORMAL,0,0) != -1) {
	gr->x = x;
	gr->y = y;
	gr->nchan = nchan;
	gr->depth = depth;
	strcpy(gr->schan,schan);
	strncpy(gr->com,com,GR_LENCOM-1);
	gr->com[GR_LENCOM-1] = '\0';
	
	return (gr);
      }
      fprintf(stderr,"NO LAYCREA\n");
      lay_destroy(&ly);
    }
    fprintf(stderr,"\n");
    free(gr);
  }
  return NULL;

}

#define GR_PNM_TYCOL 1
gr_layer *implay_pnm(FILE *fp)
{
  char bf[256],tipos[4],lastcom[256],*schan;
  gr_layer *ly;
  int i,e,a, st = 0, tipo, x, y, nchan, depth;

  while (fgets(bf,256,fp) != NULL) {
    if (bf[0] == '\n')
      continue;
    if (bf[0] == '#') {
      strcpy(lastcom,bf);
      continue;
    }
    switch(st) {
    case 0:
      /* tipo di pnm */
      tipos[0] = bf[0];
      tipos[1] = bf[1];
      tipos[2] = '\0';
      break;
    case 1:
      /* x e y dell'immagine */
      for (i=0 ; bf[i] && bf[i] != ' ' && bf[i] != '\t' ; i++)
	if (bf[i] == '\0')
	  return NULL;
      bf[i] = '\0';
      x = atoi(bf);
      y = atoi(&bf[i+1]);
      if (x<1 || y<1)
	return NULL;
      break;
    case 2:
      /* valore max raggiungibile da un singolo canale */
      depth = nbit(atoi(bf));
      break;
    }
    if (st < 2)
      st++;
    else
      break;
  }
  if (st < 2)
    return (NULL);
  /*   printf("Tipo %s\nX %d\nY %d\ndepth %d\n",tipos,x,y,depth); */
  /* definizione del tipo di pnm */
  if (!strncmp(tipos,"P6",2)) {
    tipo = GR_PNM_TYCOL;
    nchan = 3;
    schan = "rgb";
  }
  else
    return (NULL);

  strcpy(bf,&lastcom[1]);
  strcpy(lastcom,bf);
  lastcom[strlen(lastcom)-1] = '\0';
  if ((ly = lay_create(x,y,nchan,schan,depth,
		       GR_TRANS_OPAC,TRUE,lastcom)) == NULL)
    return (NULL);
  /* definisce il layer come normale */
  ly->type = GR_LAYTYP_NORMAL;
  ly->active = TRUE;
  /* legge i dati dal file */
  /* printf("y %d x %d n %d\n",y,x,nchan); */
  for (i = 0 ; i < y ; i++)
    for (e = 0 ; e < x ; e++)
      for (a = 0 ; a < nchan ; a++)
	((unsigned char *)ly->dt[a][i])[e] = (unsigned char)fgetc(fp);
  return (ly);
}

void explay_pnm(FILE *fp,gr_layer *ly)
{
  int i,e,a;
  
  if (ly->nchan == 3) {
    fprintf(fp,"P6\n#%s\n%d %d\n%d\n",ly->com, ly->x, ly->y, (1<<ly->depth)-1);
    /* printf("y %d x %d n %d\n",ly->y,ly->x,ly->nchan); */
    for (i = 0 ; i < ly->y ; i++)
      for (e = 0 ; e < ly->x ; e++)
	for (a = 0 ; a < 3 ; a++)
	  fputc((int)(((unsigned char *)ly->dt[a][i])[e]),fp);	
  }
  else if (ly->nchan == 1) {
    fprintf(fp,"P5\n#%s\n%d %d\n%d\n",ly->com, ly->x, ly->y, (1<<ly->depth)-1);
    /* printf("y %d x %d n %d\n",ly->y,ly->x,ly->nchan); */
    for (i = 0 ; i < ly->y ; i++)
      for (e = 0 ; e < ly->x ; e++)
	  fputc((int)(((unsigned char *)ly->dt[0][i])[e]),fp);	
  }
}

gripho *img_fromlay(gr_layer *ly)
{
  gripho *gr;
  
  if ((gr = malloc(sizeof(gripho))) != NULL) {
    if ((gr->lays = malloc(sizeof(gr_layer *))) != NULL) {
      ly->dx = 0;
      ly->dy = 0;
      gr->lays[0] = ly;
      gr->nlay = 1;
      gr->x = ly->x;
      gr->y = ly->y;
      gr->nchan = ly->nchan;
      gr->depth = ly->depth;
      strcpy(gr->schan,ly->schan);
      strcpy(gr->com,ly->com);
      strcpy(ly->com,"background");
      return (gr);
    }
    free(gr);
  }
  return (NULL);
}

gripho *import_pnm(FILE *fp)
{
  gripho *gr;
  gr_layer *ly;

  if ((ly = implay_pnm(fp)) != NULL) {
    if ((gr = img_fromlay(ly)) != NULL) {
      return (gr);
    }
    lay_destroy(&ly);
  }
  return NULL;
}

void chan_setdef(void **dt, int x, int y, int depth, unsigned long defval)
{
  int i,e,nb;

  nb = nbyte(depth);

  if (nb == 1) 
    for (i = 0 ; i<y ; i++)
      for (e = 0 ; e<x ; e++) 
	((unsigned char *)dt[i])[e] = (unsigned char)defval;
  else if (nb == 2)
    for (i = 0 ; i<y ; i++)
      for (e = 0 ; e<x ; e++) 
	((unsigned short *)dt[i])[e] = (unsigned short)defval;
  else
    for (i = 0 ; i<y ; i++)
      for (e = 0 ; e<x ; e++) 
	((unsigned long *)dt[i])[e] = defval;
}

int lay_addchan(gr_layer *ly,char id,unsigned long defval) 
{
  int nb;

  if ((ly->dt = (void ***)realloc(ly->dt,
				sizeof(void ***)*ly->nchan+1)) != NULL) {
    if ((ly->dt[ly->nchan] = (void **)malloc(sizeof(void **) * ly->y)) 
	!= NULL) {
      nb = nbyte(ly->depth);
      if (gr_alloc_chan(&(ly->dt[ly->nchan]),nb,ly->x,ly->y)) {
	chan_setdef(ly->dt[ly->nchan],ly->x,ly->y,nb,defval);
	ly->schan[ly->nchan++] = id;
	
	return TRUE;
      }
    }
    ly->dt = (void ***)realloc(ly->dt,sizeof(void ***)*ly->nchan);
  }
  
  return FALSE;
}

int img_addchan(gripho *gr,char id,unsigned long defval) 
{
  int i;
  
  for (i = 0 ; i < gr->nlay ; i++) {
    if (!lay_addchan(gr->lays[i],id,defval)) {

      return FALSE;
    }
  }
  gr->schan[gr->nchan++] = id;
  
  return TRUE;
}

unsigned long chan_getpix(void *dt,int x,int nb)
{
  if (nb == 1) 
    return ((unsigned long)((unsigned char *)dt)[x]);
  else if (nb == 2) 
    return ((unsigned long)((unsigned short *)dt)[x]);
  else
    return ((unsigned long)((unsigned long *)dt)[x]);
}

void chan_setpix(void *dt,int x,int nb,unsigned long val)
{
  if (nb == 1) 
    ((unsigned char *)dt)[x] = (unsigned char)val;
  else if (nb == 2) 
    ((unsigned short *)dt)[x] = (unsigned short)val;
  else
    ((unsigned long *)dt)[x] = (unsigned long)val;
}

int lay_checkpx(gr_layer *ly,int x,int y) {  
  if (!ly->active)
    return FALSE;
  if (x >= ly->dx && x < (ly->dx+ly->x) &&
      y >= ly->dy && y < (ly->dy+ly->y)) {
    return TRUE;
  }
  else {
    return FALSE;
  }
}

int chan_getnumid(char *schan,char id)
  {
    char *p;

    if ((p = strchr(schan,id)))
      return (p - schan);
    else
      return (-1);
  }

gr_layer *img_flatolay(gripho *gr,char *schan,int depth)
{
  double dbf,db,dtra,ddefout;
  gr_layer *ly;
  int x,y,a,nl,nb,alphaidx,dx,dy;

  if ((ly = lay_create(gr->x,gr->y,strlen(schan),schan,depth,
		       GR_TRANS_OPAC,TRUE,"background")) != NULL) {
    dbf = (double) ((1 << gr->depth)-1);
    ddefout = (double) ((1 << depth)-1);
    nb = nbyte(gr->depth);
    /* se esiste un canale alfa */
    alphaidx = chan_getnumid(gr->schan,GR_C_ALPHA);
    for (a = 0 ; a < gr->nchan ; a++) {
      if (alphaidx == a)
	continue;
      for (y = 0 ; y < gr->y ; y++)
	for (x = 0 ; x < gr->x ; x++) {
	  db = dbf;
	  for (nl = 0 ; nl < gr->nlay ; nl++) {
	    if (!lay_checkpx(gr->lays[nl],x,y))
	      continue;
	    dx = gr->lays[nl]->dx;
	    dy = gr->lays[nl]->dy;
	    /*	    if (x%100 == 0 && y%100 == 0)
	      fprintf(stderr,"nl %d\n",nl);*/
	    if (gr->lays[nl]->type == GR_LAYTYP_NORMAL) {

	      if (alphaidx >= 0) {
		dtra = ((double)
			(gr->lays[nl]->trans * 
			 (double)chan_getpix(gr->lays[nl]->dt[alphaidx][y-dy],x-dx,nb)) / 
			(double)(GR_TRANS_OPAC * GR_TRANS_OPAC)); 
	      }
	      else
		dtra = ((double)gr->lays[nl]->trans / 
			(double)(GR_TRANS_OPAC));
	      db = db * (1.0 - dtra) +
		(double)chan_getpix(gr->lays[nl]->dt[a][y-dy],x-dx,nb) * dtra;
	    }
	  }
	  chan_setpix(ly->dt[a][y],x,nb,(unsigned long)db);
	}
    }
  return ly;
  }
  return NULL;
}
