/*
 *   gripho.h
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <math.h>

#include "gr_geom.h"

/* -- DEFINE -- */
#define TRUE             1
#define FALSE            0
#define GR_LENCOM      128

/* tipo di canali */
#define GR_LAYTYP_NORMAL 1

/* opacita' massima */
#define GR_TRANS_TRAN    0
#define GR_TRANS_OPAC  255

/* id dei canali */
#define GR_C_RED       'r'
#define GR_C_GREEN     'g'
#define GR_C_BLUE      'b'
#define GR_C_ALPHA     'A'

/* -- DEBUGGING -- */
#include "mdebug.h"

/* -- STRUTTURE -- */
struct tag_gr_layer {
  int type;               /* tipo di layer                 */
  void ***dt;             /* area dati  dt[chan][raw][col] */
  int x;                  /* dimensione x                  */
  int y;                  /* dimensione y                  */
  int dx;                 /* displacement x                */
  int dy;                 /* displacement y                */
  int nchan;              /* numero di canali              */
  char schan[32];         /* descrizione canali            */
  int depth;              /* profondita di canale          */
  int trans;              /* trasparenza del layer         */
  int active;             /* active                        */
  char com[GR_LENCOM];    /* descrizione layer             */
};
typedef struct tag_gr_layer gr_layer;

struct tag_gripho {
  gr_layer **lays;
  int nlay;
  int x;
  int y;
  int nchan;
  int depth;
  char schan[32];         /* descrizione canali            */
  char com[GR_LENCOM];
};
typedef struct tag_gripho gripho;

/* -- PROTOTIPI -- */
double ATTRACT(double);
void gr_free_chan (void *** , int);
void gr_free_data (void **** , int , int);
int gr_alloc_chan (void ***, int , int , int);
void ***gr_alloc_data (int , int , int , int);
void lay_fliporiz (gr_layer *);
void lay_flipvert (gr_layer *);
int nbit (int);
int nbyte (int);
unsigned long chan_getpix(void * , int , int);
void chan_setpix(void * , int , int , unsigned long);
void lay_info(gr_layer *);
int lay_destroy(gr_layer **);
gr_layer *lay_create(int, int, int , char *, int, int, int, char *);
void lay_copydata(gr_layer *,gr_layer *);
void lay_copyheader(gr_layer *,gr_layer *);
gr_layer *lay_copy(gr_layer *);
unsigned long chan_getpixscaled(void *** , int , int , int ,
				double , double , int);
gr_layer *lay_scalenew (gr_layer * , int , int );
int img_addlayer (gripho * , gr_layer * , int , int , int );
gripho *new_img (int , int , int , char *, int , char *);
gripho *img_fromlay(gr_layer *);
gr_layer *implay_pnm (FILE *);
void explay_pnm (FILE * , gr_layer *);
gripho *import_pnm (FILE *);
void chan_setdef (void **, int , int , int , unsigned long);
int lay_addchan (gr_layer * , char , unsigned long);
int img_addchan (gripho * , char , unsigned long);
unsigned long chan_getpix (void * , int , int);
void chan_setpix (void * , int , int , unsigned long);
int lay_checkpx (gr_layer * , int , int);
int chan_getnumid (char * , char);
gr_layer *img_flatolay (gripho * , char * , int);
int img_lastlay(gripho *);
int img_remlayer(gripho *, int);
