/*
 *   iperb.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gripho.h"

void usage(char *s) 
{
  fprintf(stderr,"\n  Usage:\n");
  fprintf(stderr,"    %s <img1.pnm> <img2.pnm> <nframes> <colback> <middle> <prefixout>\n",s);
  fprintf(stderr,"        <colback> = rrr,ggg,bbb,AAA\n\n");
}

void 
dapac(double *c,double *p)
{
  c[0] = p[0];
  c[1] = 3.0 * (p[1] - p[0]);
  c[2] = 3.0 * (p[2] - p[0]) - 2.0 * c[1];
  c[3] = p[3] - (c[0] + c[1] + c[2]);
}

double ipsilon(double *cx, double *cy, double xok)
{
  double x,y, cur = .5, stp = .5;
  int i;

  for (i = 0 ; i<1000 ; i++) {
    x = cx[0] + (cx[1] + (cx[2] + cx[3] * cur) * cur) * cur;
    if (abs (x - xok) < 0.01)
      break;
    stp /= 2;
    if (x < xok) 
      cur += stp;
    else
      cur -= stp;
  }
  if (i == 1000)
    exit(21);
  y = cy[0] + (cy[1] + (cy[2] + (cy[3] * cur)) * cur) * cur;

  return (y);
}
int main(int argc,char **argv)
{
  gripho *gr;
  char bf[256];
  FILE *fp;
  double xx,yy,px[4],py[4],cx[4],cy[4];
  int tut,x,y,ym,c;
  int col[3] = { 0x77, 0xff, 0x77 };
  int blk[3] = { 0x54, 0x82, 0x5f };
  if ((gr = new_img(310,310,3,"rgb",8,"base")) == NULL)
    exit (1);

  px[0] = 0.0; py[0] = 310.0;
  px[3] = 310.0; py[3] = 0.0;            
  for (tut = 0 ; tut < 30 ; tut++) {
    if (tut < 5) {
      px[1] = 0.0; py[1] = -(5-tut)*15.5;
      px[2] = -(5-tut)*15.5; py[2] = 0.0;
    }
    else if (tut >=5 && tut < 25) {
      tut -= 5;
      px[1] = 0.0+(tut*15.5); py[1] = 0.0+(tut*15.5);
      px[2] = 0.0+(tut*15.5); py[2] = 0.0+(tut*15.5);
      tut += 5;
    }
    else {
      px[1] = (tut-5)*15.5; py[1] = 310.0;
      px[2] = 310.0;    py[2] = (tut-5)*15.5;
    }
    dapac(cx,px);
    dapac(cy,py);

    for (x = 0 ; x < 310 ; x++) 
      for (y = 0 ; y < 310 ; y++)
	for (c = 0 ; c<3 ; c++) {
	chan_setpix(gr->lays[0]->dt[c][y],x,1,col[c]);
	}

    for (x = 0 ; x < 310 ; x++) {
      xx = (double)x;
      yy = ipsilon(cx,cy,xx);
      ym = (int)yy;
      if (ym < 0)
	ym = 0;
      else if(ym >= 310)
	ym = 309;
      for (y = 0 ; y < ym ; y++) {
	for (c = 0 ; c<3 ; c++) {
	  chan_setpix(gr->lays[0]->dt[c][y],x,1,blk[c]);
	}
      }
    }
    sprintf(bf,"iperc%02d.pnm",tut);
    if ((fp = fopen(bf,"w")) == NULL)
      exit (2);
    explay_pnm(fp,gr->lays[0]);
    fclose(fp);
  }

#ifdef ZAGOR


  gr_layer *ly1,*ly2,*lys1,*lys2,*layer,*lytras,*lystras;

  FILE *fin1,*fin2,*fot;
  char nomin1[512],nomin2[512],prefix[512],nomeout[512];
  int i,c,x,y,nframes,col[4],dx,nb;
  int idx1,idx2,idx3;
  double delta,perc,dmax,middle;
  



  if (argc != 7) {
    usage(argv[0]);
    exit(1);
  }
  /* parte assimilazione parametri */
  strcpy(nomin1,argv[1]);
  strcpy(nomin2,argv[2]);
  if ((nframes = atoi(argv[3])) < 2) {
    fprintf(stderr,"Wrong frames number\n");
    exit(2);
  }
  sscanf(argv[4],"%d,%d,%d,%d",&col[0],&col[1],&col[2],&col[3]);
  middle = atof(argv[5]);
  if (middle <= 0.0 || middle >= 1.0) {
    fprintf(stderr,"Wrong middle value (0.0,1.0)\n");
    exit(2);
  }
  fx3d(-1,middle);
  strcpy(prefix,argv[6]);

  if ((fin1 = fopen(nomin1,"r")) == NULL) {
    fprintf(stderr,"%s open failed\n",nomin1);
    exit(3);
  }
  if ((fin2 = fopen(nomin2,"r")) == NULL) {
    fprintf(stderr,"%s open failed\n",nomin2);
    exit(4);
  }
  /* apertura dei file e check di congruenza */
  if ((ly1 = implay_pnm(fin1)) == NULL) {
    fprintf(stderr,"%s open failed\n",nomin1);
    exit(5);
  }
  if ((ly2 = implay_pnm(fin2)) == NULL) {
    fprintf(stderr,"%s open failed\n",nomin2);
    exit(6);
  }
  if (ly1->x != ly2->x || ly1->y != ly2->y) {
    fprintf(stderr,"wrong images dimensions\n");
    exit(7);
  }
  /* calcola lo step di contrazione dell'immagine */
  delta = (double)ly1->x / (double)(nframes-1);

  /* aggiunge il chan alpha ai due layer */
  lay_addchan(ly1,GR_C_ALPHA,255);
  lay_addchan(ly2,GR_C_ALPHA,255);
  /* genera l'immagine */
  if ((gr = img_fromlay(ly1)) == NULL)
    exit(8);
  nb = nbyte(gr->depth);
  lytras = lay_copy(ly1);
  for (c = 0 ; c < 4 ; c++)
    for (y = 0 ; y < gr->y ; y++) 
      for (x = 0 ; x < gr->x ; x++) 
	chan_setpix(lytras->dt[c][y],x,nb,col[c]);
  

  for (i = 0 ; i<nframes ; i++) {    
    lystras = lay_copy(lytras);
    dx = (int)((double)i * delta);
    if (i == 0) {
      lys1 = lay_copy(ly2);
      if ((idx1 = img_addlayer(gr,lys1,GR_LAYTYP_NORMAL,0,0)) == -1)
	exit(11);
      idx2 = -1;
    }
    else if (i == nframes - 1) {
      lys1 = lay_copy(ly1);
      if ((idx1 = img_addlayer(gr,lys1,GR_LAYTYP_NORMAL,0,0)) == -1)
	exit(12);
      idx2 = -1;
    }
    else {
      if ((lys1 = lay_scalenew(ly1,dx,ly1->y)) == NULL)
	exit(9);
      if ((idx1 = img_addlayer(gr,lys1,GR_LAYTYP_NORMAL,0,0)) == -1)
	exit(11);
      if ((lys2 = lay_scalenew(ly2,ly1->x-dx,ly1->y)) == NULL)
	exit(10);
      if ((idx2 = img_addlayer(gr,lys2,GR_LAYTYP_NORMAL,dx,0)) == -1)
	exit(12);
    }
    /* aggiusta il layer trasparente */
    if (i == 0 || i == nframes - 1) {
      for (x = 0 ; x < gr->x ; x++) 
	for (y = 0 ; y < gr->y ; y++)
	  chan_setpix(lystras->dt[3][y],x,nb,0);
    }    
    else {      
      dmax = sqrt((double)(gr->x * gr->x - dx * dx));
      for (x = 0 ; x < dx ; x++) {
	perc = fx3d(((dmax * (1.0 - (double)x / (double)(dx-1))) / (double)gr->x), middle);
	for (y = 0 ; y < gr->y ; y++)
	  chan_setpix(lystras->dt[3][y],x,nb,
		      chan_getpix(lytras->dt[3][y],x,nb) * perc);
      }
      dmax = sqrt((double)(gr->x * gr->x - ((gr->x - dx) * (gr->x - dx))));
      for (x = 0 ; x < gr->x - dx ; x++) {
	perc = fx3d((dmax * ((double)x / (double)(gr->x - dx-1))) / (double)gr->x,middle);
	for (y = 0 ; y < gr->y ; y++)
	  chan_setpix(lystras->dt[3][y],x+dx,nb,
		      chan_getpix(lytras->dt[3][y],x+dx,nb) * perc);
      }
    }
    if ((idx3 = img_addlayer(gr,lystras,GR_LAYTYP_NORMAL,0,0)) == -1)
      exit(13);
    sprintf(nomeout,"%s%03d.pnm",prefix,i);
    printf("out %s\n",nomeout);
    if ((fot = fopen(nomeout,"w")) == NULL) {
      fprintf(stderr,"%s open failed\n",nomeout);
      exit(13);
    }    
    /* export dell'immagine */
    if ((layer = img_flatolay(gr,"rgb",8)) == NULL)    
      fprintf(stderr,"FALLITO FLATTING\n");
    explay_pnm(fot,layer);
    fclose(fot);

    if (idx3 > -1)
      img_remlayer(gr,idx3);
    if (idx2 > -1)
      img_remlayer(gr,idx2);
    if (idx1 > -1)
      img_remlayer(gr,idx1);
  }
#endif
  exit(0);
}

