/*
 *   sharpy.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "gripho.h"


#define NBITPT 10
/* DEVE essere una potenza di 2 */
#define NPT (1<<NBITPT)

/* -1 non analizzato 
    0 chiuso */
struct gr_pt_tag { 
  int x;
  int y;
};
typedef struct gr_pt_tag gr_pt;

struct gr_crom_tag { 
  long c[3];
};
struct gr_cromdob_tag { 
  double c[3];
};
typedef struct gr_crom_tag gr_crom;
typedef struct gr_cromdob_tag gr_cromdob;

struct gr_bitplane_tag {
   int x;
   int y;
   int l;
   unsigned char *data;
};
typedef struct gr_bitplane_tag gr_bitplane;

/*
 *  VARIABILI GLOBALI 
 */

int nb;

gr_bitplane *createbitplane(int x, int y)
{
  int xx;
  gr_bitplane *new;

  xx = (x % 8 == 0 ? x / 8 : (x / 8) + 1);
  if ((new = (gr_bitplane *)malloc(sizeof(gr_bitplane))) != NULL) {
    if ((new->data = (unsigned char *)malloc(xx * y)) != NULL) {
      memset(new->data,0,xx * y);
      new->x = x;
      new->y = y;
      new->l = xx;
      return (new);
    }
    else
      free(new);
  }
  return (NULL);
}

int getbit(gr_bitplane *bp, int x, int y)
{
  int a,bit,d;

  a = x >> 3;
  bit = x & 0x07;
  d = 0x01 << bit;
  
  return ((int)((bp->data[bp->l * y + a] & d) >> bit));
}

void setbit(gr_bitplane *bp, int x, int y)
{
  int a;

  a = x >> 3;

  bp->data[bp->l * y + a] |=    0x01 << (x & 0x07);
}

void resetbit(gr_bitplane *bp, int x, int y)
{
  int a;

  a = x >> 3;

  bp->data[bp->l * y + a] &= ~(0x01 << (x & 0x07));
}


int processpix(gr_layer *ly,gr_layer *lot,gr_layer *ltr,int x,int y,
	       gr_crom *cin,gr_crom *cot,gr_cromdob *ccomp,
	       double hard,double cutoff,double soft)
  {
    double mintot = 100000000.0,tot,pretot,c[3];
    int i,e,mini,pixelbuono = 0;

    pretot = 0;
    for (e = 0 ; e < 3 ; e++) {
      c[e] = (double)chan_getpix(ly->dt[e][y],x,nb);
      pretot += (c[e] - (double)cin->c[e]) * (c[e] - (double)cin->c[e]);
    }
    
    for (i = 0 ; i < 256 ; i++) {
      tot = 0;
      for (e = 0 ; e < 3 ; e++) 
	tot += (c[e] - ccomp[i].c[e]) * (c[e] - ccomp[i].c[e]);
      if (tot < mintot) {
	mintot = tot;
	mini = i;
      }
    }
#ifdef ZAGURRO
    fprintf(stderr,"pre %f tot %f mini %d\n",pretot,mintot,mini);
    for (e = 0 ; e<3 ; e++)
      fprintf(stderr,"MINTOT %f mini %d c[e] %f ccomp[i].c[e] %f\n",
	      mintot,mini,c[e],ccomp[mini].c[e]);
    exit(1);
#endif
    if (tot < soft*soft) {
      chan_setpix(ltr->dt[0][y],x,nb,mini);
      for (e = 0; e < 3 ; e++)
	chan_setpix(lot->dt[e][y],x,nb,cot->c[e]);
      if (pretot < cutoff*cutoff)
	mini = 0;
      chan_setpix(ltr->dt[0][y],x,nb,mini);      
      if (pretot < hard*hard) 
	pixelbuono = 1;
    }

    return (pixelbuono);
  }

int checkhex(char *s)
{
  int i;
  
  for (i = 0 ; s[i] != '\0' ; i++) {
    if (s[i] < '0' || (s[i] > '9' && s[i] < 'A') ||
	(s[i] > 'F' && s[i] < 'a') || s[i] > 'f')
      return (0);
  }
  return (1);
}

int darg2crom(gr_crom *cl,char *s)
{
   int i;
   char *p = s, *t = s;

   for (i = 0 ; i < 3 ; i++) {
     for (;*t != '\0' && *t != ':' ; t++);
     *t++ = '\0';
     if (!strncmp(p,"0x",2) || !strncmp(p,"0X",2)) {
       /* gestione hex */
       if (!checkhex(&(p[2])))
	 return (0);
       cl->c[i] = strtol(&(p[2]),NULL,16);
     }
     else if (strchr(p,'.')) {
       /* gestione scientifica */
       cl->c[i] = (long)(255.0 * atof(p));
     }
     else {
       /* gestione decimale */
       cl->c[i] = atol(p);
     }
     if (cl->c[i] < 0)
       cl->c[i] = 0;
     if (cl->c[i] > 255)
       cl->c[i] = 255;     
     p = t;
   }
   return (1);
}

gr_pt pdel[] = { {  1, 0 },  { 1, 1 }, { 0, 1 }, { -1, 1 }, 
                 { -1, 0 }, { -1,-1 }, { 0,-1 }, {  1,-1 } };

/*
 *  MAIN
 */

void usage (char *s,int ext) {
  printf("%s <x> <y> <bgcol> <limcol> <threshold> <cutoff> <fartrasp> <filein> <prefileout>\n",s);
  exit(ext);
}

int main(int argc,char **argv)
{
  FILE *fin, *fot, *fot2;
  gr_layer *ly, *lot, *ltr;
  gr_bitplane *dead;
  gr_pt *spy;
  long cur = 0, icur;
  long num,nexnum,i,e,gener,ifre, curzero, fre = 1, oldfre;
  int px,py,xz,yz,x,y,pixelbuono = 0;
  char nomout[512],nomout2[512];
  double threshold, cutoff, fartrasp;
  gr_cromdob ccomp[256];
  gr_crom cin;
  gr_crom cot;

  if (argc != 10) 
    usage(argv[0],1);

  /* -- GESTIONE PARAMETRI -- */
  /* lettura pixel di partenza */
  px = atoi(argv[1]);
  py = atoi(argv[2]);  
  /* colori bg e limit */
  if (!darg2crom(&cin,argv[3]) || !darg2crom(&cot,argv[4]))
    usage(argv[0],2);
  /* soglie di intrusione, di cutoff e di manipolazione */
  for (i = 5 ; i <= 7 ; i++)
    if (atof(argv[i]) < 1.0 || atof(argv[i]) > 100000000.0)
      usage(argv[0],3);
  threshold = atof(argv[5]);
  cutoff = atof(argv[6]);
  fartrasp = atof(argv[7]);
  
  /* apertura file di input */
  if ((fin = fopen(argv[8],"r")) == NULL)
    usage(argv[0],4);
  if ((ly = implay_pnm(fin)) == NULL)
    exit(5);
  if (px < 0 || px > (ly->x-1) || 
      py < 0 || py > (ly->y-1))
    exit(6);

  sprintf(nomout,"%s-img.pnm",argv[9]);
  sprintf(nomout2,"%s-msk.pnm",argv[9]);
  if ((fot = fopen(nomout,"w")) == NULL)
    exit(7);
  if ((fot2 = fopen(nomout2,"w")) == NULL)
    exit(8);

  for (i = 0 ; i<256 ; i++) {
    for (e = 0 ; e< 3 ; e++) {
      ccomp[i].c[e] =  
	((((double)cin.c[e]) * ((double)(255-i)) / 255.0) +
	 (((double)cot.c[e]) * ((double)i) / 255.0));
    }
  }

  nb = nbyte(ly->depth);
  if ((lot = lay_copy(ly)) == NULL)
    exit(9);
  if ((ltr = lay_create(ly->x,ly->y,1,"w",8,1,1,"Gripho Mask")) == NULL)
    exit(10);
  for (i = 0 ; i< ly->x ; i++) 
    for (e = 0 ; e < ly->y ; e++)
      chan_setpix(ltr->dt[0][e],i,nb,255);

  if ((dead = createbitplane(ly->x,ly->y)) == NULL)
    exit(11);
  if ((spy = (gr_pt *)malloc(sizeof(gr_pt) * NPT)) == NULL)
    exit(12);

  spy[0].x = px;
  spy[0].y = py;
  num = 1;
  setbit(dead,px,py);
  gener = 0;
  do {
    /*    fprintf(stderr,"Gener: %ld    Spie: %ld\n",gener,num);*/

    if (gener == 0)
       processpix(ly,lot,ltr,spy[0].x,spy[0].y,
                            &cin,&cot,ccomp,threshold,cutoff,fartrasp);
    oldfre = fre;
    nexnum = 0;
    curzero = cur;
    for (; cur < curzero + num ; cur++) {
      icur = cur & (NPT-1);
      xz = spy[icur].x;
      yz = spy[icur].y;

      /* -- analisi degli otto punti intorno -- */
      for (i = 0 ; i < 7 ; i++) {
	x = xz + pdel[i].x;
	y = yz + pdel[i].y;
	/* check valid. coord. */
        if (x < 0 || x >= ly->x ||
            y < 0 || y >= ly->y)
	  continue;
	/* check per verifica stato pixel */
	if (getbit(dead,x,y))
	  continue;
	/* pixel buono, comparazione e azione seguente */
	pixelbuono = 0;
        /*
         *  QUI CI VA L'ANALISI DEL PIXEL
         */
	pixelbuono = processpix(ly,lot,ltr,x,y,
				&cin,&cot,ccomp,threshold,cutoff,fartrasp);
	
      /* -- SE QUI TESTO TROVO LA DIST. MIGLIORE TRA LE COMPOSIZIONI
            DEI DUE COLORI CONFINE -- */
        if (pixelbuono) {
	  ifre = fre & (NPT-1);
	  spy[ifre].x = x;
	  spy[ifre].y = y;
	  fre++;
	  nexnum++;
	}
	setbit(dead,x,y);
      }
    }
    
    num = nexnum;
    gener++;
  } while (oldfre != fre);
  
  explay_pnm(fot,lot);
  explay_pnm(fot2,ltr);

  fclose(fot);
  fclose(fot2);
  
  exit(0);
  return (0);
}




/*


  double bgp = .1, fgp = .9;
  tronco **trv = NULL, **tro = NULL;
  int i,ntrv,ntro;
  int bg[3] = { 0xff,0xff,0xff };
  int fg[3] = { 0x00,0x00,0x00 };


*/


