/*
 *   sharpy.c
 *   gripho Vers. 0.4
 *                 Copyright (C)  2001  Matteo Nastasi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    matteo.nastasi@milug.org
 *                 mnastasi@lycosmail.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "gripho.h"


#define NBITPT 10
/* DEVE essere una potenza di 2 */
#define NPT (1<<NBITPT)

/* -1 non analizzato 
    0 chiuso */
struct gr_pt_tag { 
  int x;
  int y;
};
typedef struct gr_pt_tag gr_pt;

struct gr_crom_tag { 
  long c[3];
};
struct gr_cromdob_tag { 
  long c[3];
};
typedef struct gr_crom_tag gr_crom;
typedef struct gr_cromdob_tag gr_cromdob;

struct gr_bitplane_tag {
   int x;
   int y;
   int l;
   unsigned char *data;
};
typedef struct gr_bitplane_tag gr_bitplane;

/*
 *  VARIABILI GLOBALI 
 */

int nb;


void usage(char *s) 
{
  fprintf(stderr,"\n  Usage:\n");
  fprintf(stderr,"    %s <img1.pnm> <img2.pnm> <nframes> <colback> <middle> <prefixout>\n",s);
  fprintf(stderr,"        <colback> = rrr,ggg,bbb,AAA\n\n");
}

gr_bitplane *createbitplane(int x, int y)
{
  int xx;
  gr_bitplane *new;

  xx = (x % 8 == 0 ? x / 8 : (x / 8) + 1);
  if ((new = (gr_bitplane *)malloc(sizeof(gr_bitplane))) != NULL) {
    if ((new->data = (unsigned char *)malloc(xx * y)) != NULL) {
      memset(new->data,0,xx * y);
      new->x = x;
      new->y = y;
      new->l = xx;
      return (new);
    }
    else
      free(new);
  }
  return (NULL);
}

int getbit(gr_bitplane *bp, int x, int y)
{
  int a,bit,d;

  a = x >> 3;
  bit = x & 0x07;
  d = 0x01 << bit;
  
  return ((int)((bp->data[bp->l * y + a] & d) >> bit));
}

void setbit(gr_bitplane *bp, int x, int y)
{
  int a;

  a = x >> 3;

  bp->data[bp->l * y + a] |=    0x01 << (x & 0x07);
}

void resetbit(gr_bitplane *bp, int x, int y)
{
  int a;

  a = x >> 3;

  bp->data[bp->l * y + a] &= ~(0x01 << (x & 0x07));
}


int processpix(gr_layer *ly,gr_layer *lot,gr_layer *ltr,int x,int y,
	       gr_crom *cin,gr_crom *cot,gr_cromdob *ccomp,
	       double hard,double soft)
  {
    double mintot = 100000000.0,tot,pretot,c[3];
    int i,e,mini,pixelbuono = 0;

    pretot = 0;
    for (e = 0 ; e < 3 ; e++) {
      c[e] = (double)chan_getpix(ly->dt[e][y],x,nb);
      pretot += (c[e] - (double)cin->c[e]) * (c[e] - (double)cin->c[e]);
    }
    
    for (i = 0 ; i < 256 ; i++) {
      tot = 0;
      for (e = 0 ; e < 3 ; e++) 
	tot += (c[e] - ccomp[i].c[e]) * (c[e] - ccomp[i].c[e]);
      if (tot < mintot) {
	mintot = tot;
	mini = i;
      }
    }
#ifdef ZAGURRO
    fprintf(stderr,"pre %f tot %f mini %d\n",pretot,mintot,mini);
    for (e = 0 ; e<3 ; e++)
      fprintf(stderr,"MINTOT %f mini %d c[e] %f ccomp[i].c[e] %f\n",
	      mintot,mini,c[e],ccomp[mini].c[e]);
    exit(1);
#endif
    if (tot < soft*soft) {
      chan_setpix(ltr->dt[0][y],x,nb,mini);
      for (e = 0; e < 3 ; e++)
	chan_setpix(lot->dt[e][y],x,nb,cot->c[e]);
      chan_setpix(ltr->dt[0][y],x,nb,mini);      
      if (pretot < hard*hard) 
	pixelbuono = 1;
    }

    return (pixelbuono);
  }


gr_pt pdel[] = { {  1, 0 },  { 1, 1 }, { 0, 1 }, { -1, 1 }, 
                 { -1, 0 }, { -1,-1 }, { 0,-1 }, {  1,-1 } };

/*
 *  MAIN
 */

int main(int argc,char **argv)
{
  FILE *fin, *fot;
  gr_layer *ly, *lot, *ltr;
  gr_bitplane *dead;
  gr_pt *spy;
  long cur = 0, icur;
  long num,nexnum,i,e,a,gener,ifre, curzero, fre = 1, oldfre;
  int px,py,xz,yz,x,y,pixelbuono = 0;
  int ngen;
  char nomout[512],nomout2[512];
  gr_cromdob ccomp[256];
  gr_crom cin = { { 255,255,255 } };
  gr_crom cot = { { 0,0,0 } };
  fin = stdin;

  for (i = 0 ; i<256 ; i++) {
    for (e = 0 ; e< 3 ; e++) {
      ccomp[i].c[e] =  
	((((double)cin.c[e]) * ((double)(255-i)) / 255.0) +
	 (((double)cot.c[e]) * ((double)i) / 255.0));
    }
  }

  px = atoi(argv[1]);
  py = atoi(argv[2]);  
  ngen = atoi(argv[3]);
  sprintf(nomout,"%s.rgba",argv[4]);
  sprintf(nomout2,"%s_msk.pnm",argv[4]);

  if ((fot = fopen(nomout,"w")) == NULL)
    exit(1);
  if ((ly = implay_pnm(fin)) == NULL)
    exit(1);
  nb = nbyte(ly->depth);
  if ((lot = lay_copy(ly)) == NULL)
    exit(1);
  if ((ltr = lay_create(ly->x,ly->y,1,"w",8,1,1,"Gripho Mask")) == NULL)
    exit(8);
#ifndef ZAGURRO
  for (i = 0 ; i< ly->x ; i++) 
    for (e = 0 ; e < ly->y ; e++)
      chan_setpix(ltr->dt[0][e],i,nb,255);
#endif

  if ((dead = createbitplane(ly->x,ly->y)) == NULL)
    exit(2);
  if ((spy = (gr_pt *)malloc(sizeof(gr_pt) * NPT)) == NULL)
    exit(3);

  spy[0].x = px;
  spy[0].y = py;
  num = 1;
  setbit(dead,px,py);
  gener = 0;
  do {
    /*    fprintf(stderr,"Gener: %ld    Spie: %ld\n",gener,num);*/

    oldfre = fre;
    nexnum = 0;
    curzero = cur;
    for (; cur < curzero + num ; cur++) {
      icur = cur & (NPT-1);
      xz = spy[icur].x;
      yz = spy[icur].y;

      /* -- analisi degli otto punti intorno -- */
      for (i = 0 ; i < 7 ; i++) {
	x = xz + pdel[i].x;
	y = yz + pdel[i].y;
	/* check valid. coord. */
        if (x < 0 || x >= ly->x ||
            y < 0 || y >= ly->y)
	  continue;
	/* check per verifica stato pixel */
	if (getbit(dead,x,y))
	  continue;
	/* pixel buono, comparazione e azione seguente */
	pixelbuono = 0;
        /*
         *  QUI CI VA L'ANALISI DEL PIXEL
         */
	pixelbuono = processpix(ly,lot,ltr,x,y,
				&cin,&cot,ccomp,2,10000);
	
      /* -- SE QUI TESTO TROVO LA DIST. MIGLIORE TRA LE COMPOSIZIONI
            DEI DUE COLORI CONFINE -- */
        if (pixelbuono) {
	  ifre = fre & (NPT-1);
	  spy[ifre].x = x;
	  spy[ifre].y = y;
	  fre++;
	  nexnum++;
	}
	setbit(dead,x,y);
      }
    }
    
    num = nexnum;
    gener++;
  } while (oldfre != fre && gener < ngen);
  
  lay_addchan(lot,GR_C_ALPHA,255);
  for (i = 0 ; i< ly->x ; i++) 
    for (e = 0 ; e < ly->y ; e++)
      ((unsigned char *)lot->dt[3][e])[i] = ((unsigned char *)ltr->dt[0][e])[i];
  
  for (e = 0 ; e < ly->y ; e++)
    for (i = 0 ; i< ly->x ; i++) 
      for (a = 0 ; a < 4 ; a++) {
	if (a == 3)
           fputc(255-(int)((unsigned char *)lot->dt[a][e])[i],fot);
        else
   	   fputc((int)((unsigned char *)lot->dt[a][e])[i],fot);
      }
  fclose(fot);
  
  exit(0);
  return (0);
}




/*


  double bgp = .1, fgp = .9;
  tronco **trv = NULL, **tro = NULL;
  int i,ntrv,ntro;
  int bg[3] = { 0xff,0xff,0xff };
  int fg[3] = { 0x00,0x00,0x00 };


*/


